# Module: PySDMs/internal
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 8/9/21
# https://github.com/daniel-furman/PySDMs

import pandas as pd
import numpy as np
import sklearn
from matplotlib import pyplot as plt, style
from pycaret import classification as pycaret
import pickle as pk

def internal_validation_visuals(min_seed, max_seed, pycaret_outdir, species_name):

      """Data vizualizations for the user-defined scoring metric (box+whisker)
      and ROC AUC visuals (10-fold stratified Cross Val) (see example
      notebooks in the Git Repo).

      min_seed/max_seed: int
          The min and max seed model runs to grab for the F1 boxplot.

      AUC_seed: int
          The seed(s) to perform ROC analysis."""
      

      # ####################################################################
      # F1 Validation Scores Data IO

      validation_scores_ensemble, validation_box_plots = [], []
      for i in np.arange(min_seed, max_seed).tolist():
          validation_scores_ensemble.append(pd.read_csv(pycaret_outdir +
          'holdout_' + str(i) + '.csv', index_col = 'Unnamed: 0')['0'])
      validation_scores_ensemble = pd.DataFrame(validation_scores_ensemble)
      for i in np.arange(0,len(list(validation_scores_ensemble))):
          validation_box_plots.append(validation_scores_ensemble[i])
      validation_scores_individual = []
      for i in np.arange(0, len(validation_scores_ensemble[0])):
          validation_scores_individual.append(np.max(
              validation_scores_ensemble.iloc[i,:]))

      # ####################################################################
      # Metric Score Hold-Out Set BoxPlots

      style.use('ggplot')
      plt.rcParams["figure.figsize"] = (2.25, 5.25)
      plt.boxplot(validation_scores_individual, 'k+', 'k^',
          medianprops = dict(color='black'), labels=['final_voter'])
      plt.title('Validation-Set self.metric Scores')
      plt.ylabel('self.metric')
      y = validation_scores_individual
      x = np.random.normal(1, 0.030, size=len(y))
      plt.plot(x, y, 'r.', alpha=0.35, markersize=11.5)
      plt.savefig(pycaret_outdir + 'metric_scores.png', dpi=400)
      plt.show()

      # ####################################################################
      # AUC K-fold Cross Validation ROC Plots

      style.use('default')
      AUC_seed=np.arange(min_seed, max_seed).tolist()
      for AUC_seed in AUC_seed:
          # Data IO
          X = pd.read_csv('test/data/env_train/env_train_'+species_name+'_'+str(AUC_seed)+'.csv')
          y = X['pa']
          X = X.drop(['pa'], axis=1)
          X = pd.DataFrame(sklearn.preprocessing.StandardScaler(
                ).fit_transform(X), columns=list(X))
          n_samples, n_features = X.shape
          cv = sklearn.model_selection.StratifiedKFold(n_splits=10)
          with open(pycaret_outdir + species_name + '_' + str(AUC_seed) + '.pkl', 'rb') as f:
                classifier = pk.load(f)
          # Classification and ROC Analysis
          tprs, aucs = [], []
          mean_fpr = np.linspace(0, 1, 100)
          fig, ax = plt.subplots(figsize=(7.5,4.5))
          for i, (train, test) in enumerate(cv.split(X, y)):
              classifier.fit(X.iloc[train, :], y.iloc[train])
              viz = sklearn.metrics.plot_roc_curve(classifier, X.iloc[
                    test, :], y.iloc[test], alpha=0.4, lw=1, ax=ax)
              interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
              interp_tpr[0] = 0.0
              tprs.append(interp_tpr)
              aucs.append(viz.roc_auc)
          ax.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r',
                label='Chance', alpha=.8)
          mean_tpr = np.mean(tprs, axis=0)
          mean_tpr[-1] = 1.0
          mean_auc = sklearn.metrics.auc(mean_fpr, mean_tpr)
          std_auc = np.std(aucs)
          ax.plot(mean_fpr, mean_tpr, color='b',
                label=r'final_voter (AUC = %0.3f $\pm$ %0.3f)' % (
                mean_auc, std_auc), lw=2, alpha=.8)
          std_tpr = np.std(tprs, axis=0)
          tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
          tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
          ax.fill_between(mean_fpr, tprs_lower, tprs_upper, color='grey',
                alpha=.2, label=r'$\pm$ 1 std. deviation')
          ax.set(xlim=[-0.05, 1.05], ylim=[-0.05, 1.05],
                title="ROC Curve: 10-fold C-V (random stratified) for seed "+str(AUC_seed))
          ax.legend(loc="lower right")
          handles, labels = ax.get_legend_handles_labels()
          ax.legend(handles[-3:], labels[-3:])
          plt.show()



# Module: PySDMs/internal
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 8/9/21
# https://github.com/daniel-furman/PySDMs

import pandas as pd
import numpy as np
from matplotlib import pyplot as plt, style
from pycaret import classification as pycaret, internal
import pickle as pk

def internal_fit(data, test_data, seed, target, exp_name, normalize, metric, fold,
    silent, mod_list=['et', 'catboost', 'rf', 'lightgbm', 'xgboost', 'gbc'],
    pycaret_outdir='outputs/', deep_learning=False, soft_voters=True, tuning=True):

      """Semi-Auto ML classification training with PyCaret, considering tree-based
        methods, neural nets (on CPU), and two-model soft voters (exhaustive
        search). Requires a Pandas data-frame with a target and explanatory
        features. Returns the voter with the best validation
        metric performance (user-defined metric). See Pycaret.org for more
        and for customization purposes (classification module).

      Returns:
          The learner with the best validation metric performance."""

      # ####################################################################
      # PyCaret Model Fitting with BRTs and soft voters

      # Setup PyCaret environment
      exp_clf = pycaret.setup(data, target=target, test_data=test_data,
          session_id=seed, experiment_name=exp_name,
          normalize=normalize, fold=fold, log_experiment=True, verbose=False,
          fold_strategy='stratifiedkfold', silent=silent) #numeric_features=[]

      # K-fold cross validation training with PyCaret "compare_models"
      classifier_list = pycaret.compare_models(n_select=len(mod_list),
          verbose=False, include=mod_list, sort=metric)
      pycaret_training_df = pycaret.pull()
      mod_keys = pycaret_training_df.index.tolist()

      if tuning:
          #custom rf grid search values for hpo
          n_estimators = [50, 60, 70, 80]
          sqrt_features = int(np.sqrt(len(pycaret.get_config('X_train').iloc[0,:])))
          max_features = [sqrt_features, sqrt_features+1, sqrt_features+2]
          min_samples_split= [sqrt_features, sqrt_features+1, sqrt_features+2]
          rf_tuned_parameters = {'max_features': max_features, 'min_samples_split': min_samples_split, 'n_estimators': n_estimators}

          #tune rf
          rf = pycaret.create_model('rf', verbose=False)
          tuned_rf, tuner_object = pycaret.tune_model(rf, verbose=False,
                custom_grid=rf_tuned_parameters, return_tuner=True)
          classifier_list.append(tuned_rf)
          mod_keys.append('tuned_rf')

          #tune rf
          et = pycaret.create_model('et', verbose=False)
          tuned_et, tuner_object = pycaret.tune_model(et, verbose=False,
                custom_grid=rf_tuned_parameters, return_tuner=True)
          classifier_list.append(tuned_et)
          mod_keys.append('tuned_et')

          #tune catboost
          catboost = pycaret.create_model('catboost', verbose=False)
          tuned_cb, tuner_object = pycaret.tune_model(catboost, verbose=False,
                search_library='optuna', return_tuner=True)
          classifier_list.append(tuned_cb)
          mod_keys.append('tuned_cb')

      # Build soft voters with best subset selection among 2 models blends
      if soft_voters:
          scoring_pandas = pd.DataFrame(np.zeros(7).reshape(-1, 7),
                columns=['Accuracy', 'AUC', 'Recall', 'Prec.', 'F1', 'Kappa',
                'MCC'])
          soft_voting_models = []
          for j in np.arange(0, len(mod_list)):
              for i in np.arange(0, len(mod_list)):
                  if i != j:
                      soft_voter = pycaret.blend_models(estimator_list=[classifier_list[j],
                            classifier_list[i]], method='soft', verbose=False)
                      soft_voting_models.append(soft_voter)
                      pull_temp = pycaret.pull()
                      scoring_pandas = scoring_pandas.append(pull_temp.loc['Mean'])
          scoring_pandas.drop(scoring_pandas.index[0], inplace=True)
          scoring_pandas.index = np.arange(0, len(scoring_pandas['Accuracy']))
          # Grab the most accurate soft_voter
          soft_voter = soft_voting_models[scoring_pandas['F1'].idxmax()]
          classifier_list.append(soft_voter)
          mod_keys.append('soft_voter')

      # Add deep learners if deep_learning (also set normalization to True)
      if deep_learning:
          MLP_2 = internal.tunable.TunableMLPClassifier(hidden_layer_sizes=(300, 150),
              activation='relu', solver='adam', random_state=seed)
          MLP_2 = pycaret.create_model(MLP_2, verbose=False)
          MLP_3 = internal.tunable.TunableMLPClassifier(hidden_layer_sizes=(300, 150,
              100), activation='relu', solver='adam', random_state=seed)
          MLP_3 = pycaret.create_model(MLP_3, verbose=False)
          MLP_4 = internal.tunable.TunableMLPClassifier(hidden_layer_sizes=(450, 300, 150,
              100), activation='relu', solver='adam', random_state=seed)
          MLP_4 = pycaret.create_model(MLP_4, verbose=False)
          # Add another soft voter with BRT + deep learners
          soft_voter_dl = pycaret.blend_models(estimator_list=[
              classifier_list[0], MLP_2, MLP_3, MLP_4],
              method='soft', verbose=False)
          # Add to function vars
          classifier_list.append(MLP_2)
          classifier_list.append(MLP_3)
          classifier_list.append(MLP_4)
          classifier_list.append(soft_voter_dl)
          mod_keys.append('MLP_2')
          mod_keys.append('MLP_3')
          mod_keys.append('MLP_4')
          mod_keys.append('soft_voter_dl')

      # ####################################################################
      # PyCaret Model Evaluation

      # Comparing cross val and hold-out strategies
      # Evaluate the models with k-fold cv and held-out partitions
      cv_mean, names, ho_mean = [], [], []
      print('The RandomState seed for this run is ' + str(seed) + '\n')
      for i in np.arange(0, len(classifier_list)):
          # first cv:
          pycaret.create_model(classifier_list[i], verbose=False)
          scorei = pycaret.pull()
          cv_mean.append(scorei[metric].iloc[0:fold])
          names.append(mod_keys[i])
          # then validation:
          pycaret.predict_model(classifier_list[i], verbose=False)
          scorei = pycaret.pull()
          ho_mean.append(scorei[metric].loc[0])
          if not silent:
              print('    >%s: CV (mean/sd): %.4f (%.4f) | Hold-out: %.4f' %
                  (names[i], np.mean(cv_mean[i]),
                  np.std(cv_mean[i]), ho_mean[i]))

      # ####################################################################
      # Finalize

      # Write the held out validation metrics to file:
      pd.DataFrame(ho_mean).to_csv(pycaret_outdir+'holdout_'+str(
          seed)+'.csv')

      # Grab model with best validation performance
      ho_mean_np = np.array(ho_mean)
      id_argmax = ho_mean_np.argmax(axis=0)
      model_key = mod_keys[id_argmax]

      # Grab the best model
      final_model_key = model_key
      if model_key == 'soft_voter':
          final_model_key = soft_voter
      if deep_learning:
          finalizing_dictionary = {'soft_voter_dl':soft_voter_dl,
            'MLP_4':MLP_4, 'MLP_2':MLP_2, 'MLP_3':MLP_3}
          for key_iter in finalizing_dictionary:
              if model_key == key_iter:
                  final_model_key = finalizing_dictionary[key_iter]
      if tuning:
          finalizing_dictionary = {'tuned_rf':tuned_rf, 'tuned_et':tuned_et, 'tuned_cb':tuned_cb}
          for key_iter in finalizing_dictionary:
              if model_key == key_iter:
                  final_model_key = finalizing_dictionary[key_iter]

      # Refit the best model with 100% of the training data
      final_voter = pycaret.finalize_model(final_model_key)

      # Test dumping and loading of .pkl model.
      with open(pycaret_outdir+exp_name+'.pkl', 'wb') as f:
          pk.dump(final_voter, f)
      with open(pycaret_outdir+exp_name+'.pkl', 'rb') as f:
          voter_pkl = pk.load(f)

      print('     >', model_key)
      print('     >', voter_pkl)

      # Returns the best learner
      return(final_voter)



# Module: PySDMs/internal
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 8/9/21
# https://github.com/daniel-furman/PySDMs

import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import rasterio
from pyimpute import load_targets, impute
import glob
import pickle as pk
import pylab

def internal_interpolate(asc_input_dir, df_input_dir, img_output_dir, seed,
                         species_name):

      """Geo-classification for model interpolation to
        raster feature surfaces. Saves to file probabilistic and binary
        predictions.

      asc_input_dir, string
          The directory containing your raster surfaces, called in a manner
          such that glob can grab all of your training features. For
          example, if your files are saved as bclim1.asc, bclim2.asc, etc.,
          then you would include bclim*.asc at the end of the string (e.g.,
          'data/train-rasters-2.5m/bclim*.asc'').

      df_input_dir, string
          The input directory for the training dataframe, as called in
          self.fit() above for the data variable. Should include all of the
          directory path except for the random seed (end of the name) and
          the .csv extension (e.g., 'data/env_train/env_train_xv_').

      img_output_dir
          The output directory for the geo-classification results.

      seed, int
          The randomstate seed of the modeling run."""

      # ####################################################################
      # Geo-classification Data IO
      
      raster_features = sorted(glob.glob(asc_input_dir))
      print('Training features / target shape:\n')
      df = pd.read_csv(df_input_dir + str(seed) + '.csv')
      train_xs = np.array(df.iloc[:,1:len(df)])
      train_y = np.array(df.iloc[:,0])
      target_xs, raster_info = load_targets(raster_features)
      print('     >', train_xs.shape, train_y.shape)

      # ####################################################################
      # Geospatial Prediction

      with open('test/outputs/' + species_name + '_' + str(seed) + '.pkl', 'rb') as f:
          classifier=pk.load(f)
      classifier.fit(train_xs, train_y)
      impute(target_xs, classifier, raster_info, outdir=img_output_dir)
      interpolation = rasterio.open(img_output_dir + 'probability_1.tif')
      response = rasterio.open(img_output_dir + 'responses.tif')

      # Image plotter function
      def plotit(x, title, cmap="Greens"):
          plt.figure()
          plt.rcParams["figure.figsize"] = (8.2, 5.5)
          pylab.imshow(x, cmap=cmap, interpolation='nearest')
          pylab.colorbar()
          pylab.title(title, fontweight='bold')
          plt.show()

      return(plotit(interpolation.read(1),'Near-Current Interpolation'),
        plotit(response.read(1),'Near-Current Interpolation'))



# Module: PySDMs
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 8/9/21
# https://github.com/daniel-furman/PySDMs

from IPython.display import display, Markdown
from IPython import get_ipython
import re
import os

from PySDMs.internal.interpolate import internal_interpolate as internal_interpolate
from PySDMs.internal.validation_visuals import internal_validation_visuals as internal_validation_visuals
from PySDMs.internal.fit import internal_fit as internal_fit

# Markdown setup for notebook environment
#def md_formatter(md, pp, cycle):
    #pp.text(md.data)
#text_plain = get_ipython().display_formatter.formatters['text/plain']
#text_plain.for_type(Markdown, md_formatter)

class PySDMs(object):

    """
    An object-oriented Python class for semi-auto ML geo-classification
    (running on PyCaret). Compares gradient boosted tree algorithms by
    default, with options to include soft voters and NNs. Designed for
    Species Distribution Modeling applications.

    Functions
    -------

    self.fit(): Classification training with PyCaret, considering tree-based
        methods, (CPU) neural nets, and two-model soft voters (exhaustive
        search). Requires a Pandas data-frame with a target and explanatory
        features. Returns the voter with the best validation
        metric performance (user-defined metric). See Pycaret.org for more
        and for customization purposes (classification module).

    self.interpolate(): Geo-classification for model interpolation to
        raster feature surfaces. Saves to file probabilistic and binary
        predictions.

    self.validation_performance(): Data vizualizations for the user-defined
        scoring metric (box+whisker) and ROC AUC visuals (10-fold stratified
        Cross Val) (see example notebooks in the Git Repo)."""

    def __init__(self, data, test_data, seed, target, exp_name,
        normalize=True, metric='F1', fold=10, silent=False,
        mod_list = ['et', 'catboost', 'rf', 'lightgbm', 'xgboost', 'gbc'],
        pycaret_outdir='test/outputs/'):

        """data: Pandas DataFrame
            Data-frame with the classification target variable and the
            explanatory features to be included in the model.

        test_data: Pandas DataFrame
            Data-frame with the classification target variable and the
            explanatory features to be used for the validation of
            the model.

        seed: int
            The RandomState seed for the experiment.

        target: string
            The col name of the species presence vs. absence target.

        exp_name: string
            Name that encodes the experiment. A key for recovering the PyCaret
            workflow with pycaret.get_logs(exp_name).

        normalize: bool, default = True
            Whether or not to perform standardization via the z-score method
            during the modeling experiment.

        metric: string, default = 'F1'
            The metric to use for model comparison, see PyCaret.setup() for
            available keys.

        fold: int, default = 10
            The number of CV k-folds for the stratified kfold strategy.

        silent: bool, default = False
            When set to True most of the output is hidden and the PyCaret
            workflow runs without user confirmation, designed for multiple runs
            between different seeds/data samples/data partitions (see the
            examples).

        mod_list: string list, default #{BRTs}
            See PyCaret for mod strings. The initial set of models to consider
            for best subset soft voter selection.

        Returns:
            The initialized PySDMs object."""

        # Init Variable Assignment
        self.data = data
        self.test_data = test_data
        self.target = target
        self.seed = seed
        self.exp_name = exp_name
        self.normalize = normalize
        self.metric = metric
        self.fold = fold
        self.silent = silent
        self.mod_list = mod_list
        self.species_name = re.split('[_]', exp_name)[0]

        DATA = os.path.dirname(os.path.abspath(__file__))
        self.output_dir =  [os.path.join(DATA, x)
            for x in [pycaret_outdir]][0]



    def fit(self, deep_learning=False, soft_voters=False, tuning=False):
        # ML classification fitting
        #display(Markdown('--- \n ### PyCaret Model Fitting: \n --- ' ))
        return(internal_fit(self.data, self.test_data, self.seed, self.target,
            self.exp_name, self.normalize, self.metric, self.fold,
            self.silent, self.mod_list, self.output_dir, deep_learning,
            soft_voters, tuning))

    def interpolate(self, asc_input_dir, df_input_dir, img_output_dir):
        # Geospatial step for raster predictions
        #display(Markdown("--- \n ### Geo-classification Interpolation: \n ---"))
        return(internal_interpolate(asc_input_dir, df_input_dir,
            img_output_dir, self.seed, self.species_name))

    def validation_visuals(self, min_seed, max_seed, pycaret_outdir):
        # Dataviz for model fitting
        #display(Markdown('--- \n ### Model Performance Plots: \n ---'))
        return(internal_validation_visuals(min_seed, max_seed, self.output_dir, self.species_name))



# Module: PySDMs
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 4/7/21

# https://github.com/daniel-furman/PySDMs

import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="PySDMs",
    version="1.0",
    author="Daniel Ryan Furman",
    author_email="dryanfurman@gmail.com",
    description=("An object-oriented class for semi-Auto ML for geospatial classification tasks."),
    long_description="See documentation at https://github.com/daniel-furman/PySDMs",
    license="MIT",
    url="https://github.com/daniel-furman/PySDMs",
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['test/data/env_train/env_train_cr_192.csv',
                       'test/data/env_train/env_train_cr_192_metadata.csv', 
                       'test/outputs/holdout_192.csv']},
    install_requires=["numpy", "pandas", "sklearn", "matplotlib", "pycaret",
        "pyimpute", "geopandas", "rasterio"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
        ],
)



# Module: PySDMs
# Author: Daniel Ryan Furman <dryanfurman@gmail.com>
# License: MIT
# Last modified : 8.11.21

from PySDMs import PySDMs
import numpy as np
import pandas as pd
import os
from pycaret import classification as pycaret, internal


def test_PySDMs():

    #Data IO
    DATA = os.path.dirname(os.path.abspath(__file__))
    data = [
        os.path.join(DATA, x)
        for x in ['data/env_train/env_train_cr_192_metadata.csv',
                  'data/train-rasters/bclim*.tif']
    ]
    output = [
        os.path.join(DATA, x)
        for x in ['outputs/']][0]

    # Run PySDMs object over the random seed(s)
    species_code = 'cr'

    for seed in [192,]:

        # Data IO
        df_input_dir = 'test/data/env_train/env_train_'+species_code+'_'
        tif_input_dir = data[1]
        mod_list = ['rf', 'et']
        col_names = ['pa','bclim7', 'bclim4', 'bclim3', 'bclim2', 'bclim5', 'bclim6',
           'bclim11', 'bclim8', 'bclim9', 'bclim10', 'bclim15', 'bclim13',
           'bclim19', 'bclim16', 'bclim12', 'bclim17', 'bclim1', 'bclim18']

        data = pd.read_csv(data[0])
        exp_name = species_code+'_'+str(seed)
        train_data = data[(data['fold']!=3) | (data['fold']!=4)][col_names]
        test_data = data[(data['fold']==3) | (data['fold']==4)][col_names]
        print()
        print('test % of total data: ', len(test_data['pa'])/(len(train_data['pa'])+len(test_data['pa'])))

        # Initialize class
        coastal_redwoods = PySDMs(train_data, test_data, seed, target='pa',
                exp_name=species_code+'_'+str(seed), normalize=False, metric='AUC',
                silent=True, mod_list=mod_list)

        # Model Fitting with self.fit()
        learner = coastal_redwoods.fit()
        pd.Series(pycaret.get_config('X_train').columns).to_csv(output+'features_'+str(seed)+'.csv')
        assert os.path.isfile(output + 'cr_192.pkl')

        # Geo-classification with self.interpolate()
        coastal_redwoods.interpolate(tif_input_dir, df_input_dir, output)
        coastal_redwoods.validation_visuals(min_seed=seed, max_seed=seed+1, pycaret_outdir=output)

    assert os.path.isfile(output + 'probability_1.tif')
